@@ -13,11 +13,15 @@
 #    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #    License for the specific language governing permissions and limitations
 #    under the License.
-import os
-import time
-import uuid
 
-import testscenarios
+import os
+import threading
+import time
+
+from concurrent import futures
+import fixtures
+import mock
+from six.moves.urllib import parse
 from testtools import matchers
 from testtools import testcase
 
@@ -25,26 +29,18 @@ import tooz.coordination
 from tooz import tests
 
 
-class TestAPI(testscenarios.TestWithScenarios,
-              tests.TestCaseSkipNotImplemented):
+def try_to_lock_job(name, coord, url, member_id):
+    if not coord:
+        coord = tooz.coordination.get_coordinator(
+            url, member_id)
+        coord.start()
+    lock2 = coord.get_lock(name)
+    return lock2.acquire(blocking=False)
 
-    scenarios = [
-        ('kazoo', {'url': os.getenv("TOOZ_TEST_ZOOKEEPER_URL"),
-                   'bad_url': 'kazoo://localhost:1'}),
-        ('zake', {'url': 'zake://?timeout=5'}),
-        ('memcached', {'url': os.getenv("TOOZ_TEST_MEMCACHED_URL"),
-                       'bad_url': 'memcached://localhost:1',
-                       'timeout_capable': True}),
-        ('ipc', {'url': 'ipc://'}),
-        ('file', {'url': 'file:///tmp'}),
-        ('redis', {'url': os.getenv("TOOZ_TEST_REDIS_URL"),
-                   'bad_url': 'redis://localhost:1',
-                   'timeout_capable': True}),
-        ('postgresql', {'url': os.getenv("TOOZ_TEST_PGSQL_URL"),
-                        'bad_url': 'postgresql://localhost:1'}),
-        ('mysql', {'url': os.getenv("TOOZ_TEST_MYSQL_URL"),
-                   'bad_url': 'mysql://localhost:1'}),
-    ]
+
+class TestAPI(tests.TestCaseSkipNotImplemented):
+
+    url = os.getenv("TOOZ_TEST_URL")
 
     def assertRaisesAny(self, exc_classes, callable_obj, *args, **kwargs):
         checkers = [matchers.MatchesException(exc_class)
@@ -57,8 +53,9 @@ class TestAPI(testscenarios.TestWithScenarios,
         super(TestAPI, self).setUp()
         if self.url is None:
             self.skipTest("No URL set for this driver")
-        self.group_id = self._get_random_uuid()
-        self.member_id = self._get_random_uuid()
+        self.useFixture(fixtures.NestedTempfile())
+        self.group_id = tests.get_random_uuid()
+        self.member_id = tests.get_random_uuid()
         self._coord = tooz.coordination.get_coordinator(self.url,
                                                         self.member_id)
         self._coord.start()
@@ -67,11 +64,14 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.stop()
         super(TestAPI, self).tearDown()
 
-    def test_connection_error(self):
-        if not hasattr(self, "bad_url"):
-            raise testcase.TestSkipped("No bad URL provided")
-        coord = tooz.coordination.get_coordinator(self.bad_url,
-                                                  self.member_id)
+    def test_connection_error_bad_host(self):
+        if (tooz.coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS
+           not in self._coord.CHARACTERISTICS):
+            self.skipTest("This driver is not distributed across hosts")
+        scheme = parse.urlparse(self.url).scheme
+        coord = tooz.coordination.get_coordinator(
+            "%s://localhost:1/f00" % scheme,
+            self.member_id)
         self.assertRaises(tooz.coordination.ToozConnectionError,
                           coord.start)
 
@@ -86,6 +86,27 @@ class TestAPI(testscenarios.TestWithScenarios,
         all_group_ids = self._coord.get_groups().get()
         self.assertTrue(self.group_id in all_group_ids)
 
+    def test_get_lock_release_broken(self):
+        name = tests.get_random_uuid()
+        memberid2 = tests.get_random_uuid()
+        coord2 = tooz.coordination.get_coordinator(self.url,
+                                                   memberid2)
+        coord2.start()
+        lock1 = self._coord.get_lock(name)
+        lock2 = coord2.get_lock(name)
+        self.assertTrue(lock1.acquire(blocking=False))
+        self.assertFalse(lock2.acquire(blocking=False))
+        self.assertTrue(lock2.break_())
+        self.assertTrue(lock2.acquire(blocking=False))
+        self.assertFalse(lock1.release())
+        # Assert lock is not accidentally broken now
+        memberid3 = tests.get_random_uuid()
+        coord3 = tooz.coordination.get_coordinator(self.url,
+                                                   memberid3)
+        coord3.start()
+        lock3 = coord3.get_lock(name)
+        self.assertFalse(lock3.acquire(blocking=False))
+
     def test_create_group_already_exist(self):
         self._coord.create_group(self.group_id).get()
         create_group = self._coord.create_group(self.group_id)
@@ -93,12 +114,12 @@ class TestAPI(testscenarios.TestWithScenarios,
                           create_group.get)
 
     def test_get_groups(self):
-        groups_ids = [self._get_random_uuid() for _ in range(0, 5)]
+        groups_ids = [tests.get_random_uuid() for _ in range(0, 5)]
         for group_id in groups_ids:
             self._coord.create_group(group_id).get()
         created_groups = self._coord.get_groups().get()
         for group_id in groups_ids:
-            self.assertTrue(group_id in created_groups)
+            self.assertIn(group_id, created_groups)
 
     def test_delete_group(self):
         self._coord.create_group(self.group_id).get()
@@ -177,9 +198,19 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertRaises(tooz.coordination.MemberNotJoined,
                           leave_group.get)
 
+    def test_get_lock_twice_locked_one_released_two(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock2 = self._coord.get_lock(name)
+        self.assertTrue(lock1.acquire())
+        self.assertFalse(lock2.acquire(blocking=False))
+        self.assertFalse(lock2.release())
+        self.assertTrue(lock1.release())
+        self.assertFalse(lock2.release())
+
     def test_get_members(self):
-        group_id_test2 = self._get_random_uuid()
-        member_id_test2 = self._get_random_uuid()
+        group_id_test2 = tests.get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -209,6 +240,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         capa = self._coord.get_member_capabilities(self.group_id,
                                                    self.member_id).get()
         self.assertEqual(capa, caps)
+        self.assertEqual(capa['type'], caps['type'])
 
     def test_get_member_capabilities_nonexistent_group(self):
         capa = self._coord.get_member_capabilities(self.group_id,
@@ -225,6 +257,44 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertRaises(tooz.coordination.MemberNotJoined,
                           capa.get)
 
+    def test_get_member_info(self):
+        self._coord.create_group(self.group_id).get()
+        self._coord.join_group(self.group_id, b"test_capabilities")
+
+        member_info = self._coord.get_member_info(self.group_id,
+                                                  self.member_id).get()
+        self.assertEqual(member_info['capabilities'], b"test_capabilities")
+
+    def test_get_member_info_complex(self):
+        self._coord.create_group(self.group_id).get()
+        caps = {
+            'type': 'warrior',
+            'abilities': ['fight', 'flight', 'double-hit-damage'],
+        }
+        member_info = {'capabilities': 'caps',
+                       'created_at': '0',
+                       'updated_at': '0'}
+        self._coord.join_group(self.group_id, caps)
+        member_info = self._coord.get_member_info(self.group_id,
+                                                  self.member_id).get()
+        self.assertEqual(member_info['capabilities'], caps)
+
+    def test_get_member_info_nonexistent_group(self):
+        member_info = self._coord.get_member_info(self.group_id,
+                                                  self.member_id)
+        # Drivers raise one of those depending on their capability
+        self.assertRaisesAny([tooz.coordination.MemberNotJoined,
+                              tooz.coordination.GroupNotCreated],
+                             member_info.get)
+
+    def test_get_member_info_nonjoined_member(self):
+        self._coord.create_group(self.group_id).get()
+        member_id = tests.get_random_uuid()
+        member_info = self._coord.get_member_info(self.group_id,
+                                                  member_id)
+        self.assertRaises(tooz.coordination.MemberNotJoined,
+                          member_info.get)
+
     def test_update_capabilities(self):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id, b"test_capabilities1").get()
@@ -248,10 +318,30 @@ class TestAPI(testscenarios.TestWithScenarios,
                              update_cap.get)
 
     def test_heartbeat(self):
+        if not self._coord.requires_beating:
+            raise testcase.TestSkipped("Test not applicable (heartbeating"
+                                       " not required)")
         self._coord.heartbeat()
 
+    def test_heartbeat_loop(self):
+        if not self._coord.requires_beating:
+            raise testcase.TestSkipped("Test not applicable (heartbeating"
+                                       " not required)")
+
+        heart = self._coord.heart
+        self.assertFalse(heart.is_alive())
+        heart.start()
+
+        # This will timeout if nothing ever is done...
+        try:
+            while not heart.beats:
+                time.sleep(1)
+        finally:
+            heart.stop()
+            heart.wait()
+
     def test_disconnect_leave_group(self):
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -267,11 +357,20 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertTrue(member_id_test2 not in members_ids)
 
     def test_timeout(self):
-        if not getattr(self, "timeout_capable", False):
-            self.skipTest("This test only works with timeout capable drivers")
-        member_id_test2 = self._get_random_uuid()
-        client2 = tooz.coordination.get_coordinator(self.url,
-                                                    member_id_test2)
+        if (tooz.coordination.Characteristics.NON_TIMEOUT_BASED
+           in self._coord.CHARACTERISTICS):
+            self.skipTest("This driver is not based on timeout")
+        self._coord.stop()
+        if "?" in self.url:
+            sep = "&"
+        else:
+            sep = "?"
+        url = self.url + sep + "timeout=5"
+        self._coord = tooz.coordination.get_coordinator(url, self.member_id)
+        self._coord.start()
+
+        member_id_test2 = tests.get_random_uuid()
+        client2 = tooz.coordination.get_coordinator(url, member_id_test2)
         client2.start()
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
@@ -309,7 +408,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         return 42
 
     def test_watch_group_join(self):
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -343,7 +442,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertIsNone(self.event)
 
     def test_watch_leave_group(self):
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -394,7 +493,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.watch_join_group(self.group_id, self._set_event)
         self._coord.watch_leave_group(self.group_id, self._set_event)
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -430,7 +529,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.watch_join_group(self.group_id, self._set_event)
         self._coord.watch_leave_group(self.group_id, self._set_event)
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -476,7 +575,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
         self._coord.run_watchers()
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -532,7 +631,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
         self._coord.run_watchers()
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -570,57 +669,261 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertEqual(self.group_id,
                          self.event.group_id)
 
+    def test_unwatch_elected_as_leader(self):
+        # Create a group and add a elected_as_leader callback
+        self._coord.create_group(self.group_id).get()
+        self._coord.watch_elected_as_leader(self.group_id, self._set_event)
+
+        # Ensure exactly one leader election hook exists
+        self.assertEqual(1,
+                         len(self._coord._hooks_elected_leader[self.group_id]))
+
+        # Unwatch, and ensure no leader election hooks exist
+        self._coord.unwatch_elected_as_leader(self.group_id, self._set_event)
+        self.assertEqual(0, len(self._coord._hooks_elected_leader))
+
+    def test_unwatch_elected_as_leader_callback_not_found(self):
+        self._coord.create_group(self.group_id).get()
+        self.assertRaises(tooz.coordination.WatchCallbackNotFound,
+                          self._coord.unwatch_elected_as_leader,
+                          self.group_id, lambda x: None)
+
+    def test_unwatch_join_group_callback_not_found(self):
+        self._coord.create_group(self.group_id).get()
+        self.assertRaises(tooz.coordination.WatchCallbackNotFound,
+                          self._coord.unwatch_join_group,
+                          self.group_id, lambda x: None)
+
+    def test_unwatch_leave_group(self):
+        # Create a group and add a leave_group callback
+        self._coord.create_group(self.group_id).get()
+        self.assertEqual(0, len(self._coord._hooks_leave_group))
+        self._coord.watch_leave_group(self.group_id, self._set_event)
+
+        # Ensure exactly one leave group hook exists
+        self.assertEqual(1, len(self._coord._hooks_leave_group[self.group_id]))
+
+        # Unwatch, and ensure no leave group hooks exist
+        self._coord.unwatch_leave_group(self.group_id, self._set_event)
+        self.assertEqual(0, len(self._coord._hooks_leave_group))
+
+    def test_unwatch_leave_group_callback_not_found(self):
+        self._coord.create_group(self.group_id).get()
+        self.assertRaises(tooz.coordination.WatchCallbackNotFound,
+                          self._coord.unwatch_leave_group,
+                          self.group_id, lambda x: None)
+
     def test_get_lock(self):
-        lock = self._coord.get_lock(self._get_random_uuid())
-        self.assertEqual(True, lock.acquire())
-        lock.release()
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        self.assertTrue(lock.acquire())
+        self.assertTrue(lock.release())
         with lock:
             pass
 
-    def test_get_multiple_locks_with_same_coord(self):
-        name = self._get_random_uuid()
+    def test_get_lock_concurrency_locking_same_lock(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+
+        graceful_ending = threading.Event()
+
+        def thread():
+            self.assertTrue(lock.acquire())
+            self.assertTrue(lock.release())
+            graceful_ending.set()
+
+        t = threading.Thread(target=thread)
+        t.daemon = True
+        with lock:
+            t.start()
+            # Ensure the thread try to get the lock
+            time.sleep(.1)
+        t.join()
+        graceful_ending.wait(.2)
+        self.assertTrue(graceful_ending.is_set())
+
+    def _do_test_get_lock_concurrency_locking_two_lock(self, executor,
+                                                       use_same_coord):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        with lock1:
+            with executor(max_workers=1) as e:
+                coord = self._coord if use_same_coord else None
+                f = e.submit(try_to_lock_job, name, coord, self.url,
+                             tests.get_random_uuid())
+                self.assertFalse(f.result())
+
+    def test_get_lock_concurrency_locking_two_lock_process(self):
+        self._do_test_get_lock_concurrency_locking_two_lock(
+            futures.ProcessPoolExecutor, False)
+
+    def test_get_lock_concurrency_locking_two_lock_thread1(self):
+        self._do_test_get_lock_concurrency_locking_two_lock(
+            futures.ThreadPoolExecutor, False)
+
+    def test_get_lock_concurrency_locking_two_lock_thread2(self):
+        self._do_test_get_lock_concurrency_locking_two_lock(
+            futures.ThreadPoolExecutor, True)
+
+    def test_get_lock_concurrency_locking2(self):
+        # NOTE(sileht): some database based lock can have only
+        # one lock per connection, this test ensures acquiring a
+        # second lock doesn't release the first one.
+        lock1 = self._coord.get_lock(tests.get_random_uuid())
+        lock2 = self._coord.get_lock(tests.get_random_uuid())
+
+        graceful_ending = threading.Event()
+        thread_locked = threading.Event()
+
+        def thread():
+            with lock2:
+                try:
+                    self.assertFalse(lock1.acquire(blocking=False))
+                except tooz.NotImplemented:
+                    pass
+                thread_locked.set()
+            graceful_ending.set()
+
+        t = threading.Thread(target=thread)
+        t.daemon = True
+
+        with lock1:
+            t.start()
+            thread_locked.wait()
+            self.assertTrue(thread_locked.is_set())
+        t.join()
+        graceful_ending.wait()
+        self.assertTrue(graceful_ending.is_set())
+
+    def test_get_lock_twice_locked_twice(self):
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
-        self.assertEqual(True, lock1.acquire())
-        self.assertEqual(False, lock2.acquire(blocking=False))
-        self.assertEqual(False,
-                         self._coord.get_lock(name).acquire(blocking=False))
-        lock1.release()
+        with lock1:
+            self.assertFalse(lock2.acquire(blocking=False))
 
-    def test_get_multiple_locks_with_same_coord_without_ref(self):
-        # NOTE(sileht): weird test case who want a lock that can't be
-        # released ? This tests is here to ensure that the
-        # acquired first lock in not vanished by the gc and get accidentally
-        # released.
-        # This test ensures that the consumer application will stuck when it
-        # looses the ref of a acquired lock instead of create a race.
-        # Also, by its nature this tests don't cleanup the created
-        # semaphore by the ipc:// driver, don't close opened files and
-        # sql connections and that the desired behavior.
-        name = self._get_random_uuid()
-        self.assertEqual(True, self._coord.get_lock(name).acquire())
-        self.assertEqual(False,
-                         self._coord.get_lock(name).acquire(blocking=False))
+    def test_get_lock_context_fails(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock2 = self._coord.get_lock(name)
+        with mock.patch.object(lock2, 'acquire', return_value=False):
+            with lock1:
+                self.assertRaises(
+                    tooz.coordination.LockAcquireFailed,
+                    lock2.__enter__)
+
+    def test_get_lock_context_check_value(self):
+        name = tests.get_random_uuid()
+        lock = self._coord.get_lock(name)
+        with lock as returned_lock:
+            self.assertEqual(lock, returned_lock)
+
+    def test_lock_context_manager_acquire_no_argument(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock2 = self._coord.get_lock(name)
+        with lock1():
+            self.assertFalse(lock2.acquire(blocking=False))
+
+    def test_lock_context_manager_acquire_argument_return_value(self):
+        name = tests.get_random_uuid()
+        blocking_value = 10.12
+        lock = self._coord.get_lock(name)
+        with lock(blocking_value) as returned_lock:
+            self.assertEqual(lock, returned_lock)
+
+    def test_lock_context_manager_acquire_argument_release_within(self):
+        name = tests.get_random_uuid()
+        blocking_value = 10.12
+        lock = self._coord.get_lock(name)
+        with lock(blocking_value) as returned_lock:
+            self.assertTrue(returned_lock.release())
+
+    def test_lock_context_manager_acquire_argument(self):
+        name = tests.get_random_uuid()
+        blocking_value = 10.12
+        lock = self._coord.get_lock(name)
+        with mock.patch.object(lock, 'acquire', wraps=True, autospec=True) as \
+                mock_acquire:
+            with lock(blocking_value):
+                mock_acquire.assert_called_once_with(blocking_value)
+
+    def test_lock_context_manager_acquire_argument_timeout(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock2 = self._coord.get_lock(name)
+        with lock1:
+            try:
+                with lock2(False):
+                    self.fail('Lock acquire should have failed')
+            except tooz.coordination.LockAcquireFailed:
+                pass
+
+    def test_get_lock_locked_twice(self):
+        name = tests.get_random_uuid()
+        lock = self._coord.get_lock(name)
+        with lock:
+            self.assertFalse(lock.acquire(blocking=False))
+
+    def test_get_multiple_locks_with_same_coord(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock2 = self._coord.get_lock(name)
+        self.assertTrue(lock1.acquire())
+        self.assertFalse(lock2.acquire(blocking=False))
+        self.assertFalse(self._coord.get_lock(name).acquire(blocking=False))
+        self.assertTrue(lock1.release())
+
+    def test_ensure_acquire_release_return(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        self.assertTrue(lock1.acquire())
+        self.assertTrue(lock1.release())
+        self.assertFalse(lock1.release())
 
     def test_get_lock_multiple_coords(self):
-        member_id2 = self._get_random_uuid()
+        member_id2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id2)
         client2.start()
 
-        lock_name = self._get_random_uuid()
+        lock_name = tests.get_random_uuid()
         lock = self._coord.get_lock(lock_name)
-        self.assertEqual(True, lock.acquire())
+        self.assertTrue(lock.acquire())
 
         lock2 = client2.get_lock(lock_name)
-        self.assertEqual(False, lock2.acquire(blocking=False))
-        lock.release()
-        self.assertEqual(True, lock2.acquire(blocking=True))
-        lock2.release()
+        self.assertFalse(lock2.acquire(blocking=False))
+        self.assertTrue(lock.release())
+        self.assertTrue(lock2.acquire(blocking=True))
+        self.assertTrue(lock2.release())
 
-    @staticmethod
-    def _get_random_uuid():
-        return str(uuid.uuid4()).encode('ascii')
+    def test_get_started_status(self):
+        self.assertTrue(self._coord.is_started)
+        self._coord.stop()
+        self.assertFalse(self._coord.is_started)
+        self._coord.start()
+
+    def do_test_name_property(self):
+        name = tests.get_random_uuid()
+        lock = self._coord.get_lock(name)
+        self.assertEqual(name, lock.name)
+
+    def test_acquire_twice_no_deadlock_releasing(self):
+        name = tests.get_random_uuid()
+        lock = self._coord.get_lock(name)
+        self.assertTrue(lock.acquire(blocking=False))
+        self.assertFalse(lock.acquire(blocking=False))
+        self.assertTrue(lock.release())
+
+
+class ZakeTestAPI(TestAPI):
+    url = "zake://"
+
+
+class IPCTestAPI(TestAPI):
+    url = "ipc://"
+
+
+class FileTestAPI(TestAPI):
+    url = "file:///tmp"
 
 
 class TestHook(testcase.TestCase):
